<?php

// Checking PHP Version
if (version_compare(PHP_VERSION, '7.4.0', '<')) {
    die('Payment module "Payment system VTB" requires PHP version 7.4.0 or higher.');
}

class nc_payment_system_vtbpay extends nc_payment_system {
  
    /**
     * Indicates whether the payment system is automatic or not.
     *
     * @var bool
     */
    protected $automatic = TRUE;
    
    /**
     * The version of the payment system.
     *
     * @var string
     */
    public $version = '1.0.1';
    
    /**
     * List of accepted currencies.
     *
     * @var array
     */
    protected $accepted_currencies = ['RUB', 'RUR'];
    
    /**
     * Mapping of currency aliases to their canonical names.
     *
     * @var array
     */
    protected $currency_map = ['RUR' => 'RUB'];
    
    /**
     * Settings of the site in the payment system.
     * Includes client ID, client secret, merchant authorization, and various URLs.
     *
     * @var array
     */
    protected $settings = [
        'client_id' => null,
        'client_secret' => null,
        'merchant_authorization' => null,
        'test_mode' => 0,
        'logging' => 0,
        'success_url' => null,
        'fail_url' => null
    ];
    
    /**
     * Parameters to be sent to the payment system.
     *
     * @var array
     */
    protected $request_parameters = [];

    /**
     * Parameters received from the payment system.
     *
     * @var array
     */
    protected $callback_response = [
        'orderId' => null
    ];
    
    /**
     * URL for making payments.
     *
     * @var string|null
     */
    protected $pay_url = null;
    
    /**
     * URL for receiving responses from the payment system.
     *
     * @var string|null
     */
    protected $response_url = null;
    
    /**
     * Constants for log levels
     */
    const LOG_LEVEL_DEBUG = 'DEBUG';
    const LOG_LEVEL_ERROR = 'ERROR';
    
    
    /**
     * VtbApi
     */

    /**
     * VtbApi settings and endpoints.
     * 
     * @var string|null
     */
    private $tokenUrl;
    
    /**
     * URL for creating orders in the payment system.
     *
     * @var string|null
     */
    private $orderUrl;
    
    /**
     * Mode of the payment system (test or production).
     *
     * @var string|null
     */
    private $mode;
    
    /**
     * Merchant authorization data for the client.
     *
     * @var mixed
     */
    private $clientMerchant;
    
    /**
     * Data for the client, including the grant type for OAuth.
     *
     * @var array
     */
    private $clientData = [
        'grant_type' => 'client_credentials'
    ];
    
    /**
     * Mapping of modes to their corresponding URLs.
     *
     * @var array
     */
    static private $modeUrls = [
        'PROD' => [
            'TOKEN' => 'https://open.api.vtb.ru:443/passport/oauth2/token',
            'ORDER' => 'https://gw.api.vtb.ru/openapi/smb/efcp/e-commerce/v1/orders'
        ],
        'TEST' => [
            'TOKEN' => 'https://auth.bankingapi.ru/auth/realms/kubernetes/protocol/openid-connect/token',
            'ORDER' => 'https://hackaton.bankingapi.ru/api/smb/efcp/e-commerce/api/v1/orders'
        ]
    ];
    
    
    /**
     * The class constructor. Adds listeners for the specified events
     * related to payment requests and callbacks.
     */
    public function __construct() 
    {
        parent::__construct();
        
        $event = nc_core::get_object()->event;
        $event->add_listener(
            nc_payment_system::EVENT_AFTER_PAY_REQUEST, [
            $this, 
            'on_after_pay_request'
        ]);
        $event->add_listener(
            nc_payment_system::EVENT_AFTER_PAY_CALLBACK, [
            $this, 
            'on_after_pay_callback'
        ]);
        $event->add_listener(
            nc_payment_system::EVENT_ON_PAY_CALLBACK_ERROR, [
            $this, 
            'on_pay_callback_error'
        ]);
    }
    
    
    /**
     * Redirects to the payment page after a payment request is created
     *
     * @param nc_payment_system_vtbpay $this_obj Object of the current class
     * @param nc_payment_invoice $invoice The invoice for the payment
     */
    public function on_after_pay_request(
        nc_payment_system_vtbpay $this_obj, 
        nc_payment_invoice $invoice = null
    ) {
        if (!empty($this->pay_url)) {
            header('Location: ' . $this->pay_url);
        }
    }
    
    
    /**
     * Redirects to the appropriate URL based on the payment callback response
     *
     * @param nc_payment_system_vtbpay $this_obj Object of the current class
     * @param nc_payment_invoice $invoice The invoice for the payment
     */
    public function on_after_pay_callback(
        nc_payment_system_vtbpay $this_obj, 
        nc_payment_invoice $invoice = null
    ) {
        if (!empty($this->response_url)) {
            header('Location: ' . $this->response_url);
        }
    }
    
    
    /**
     * Redirects to the error URL.
     *
     * @param nc_payment_system_vtbpay $this_obj Object of the current class
     * @param nc_payment_invoice $invoice The invoice for the payment
     */
    public function on_pay_callback_error(
        nc_payment_system_vtbpay $this_obj, 
        nc_payment_invoice $invoice = null
    ) {
        if (!empty($invoice)) $this->on_payment_failure($invoice);
        header('Location: ' . $this->get_setting('fail_url') ?: '/');
    }
    
    
    /**
     * Processes the payment request and initiates the payment procedure
     *
     * @param nc_payment_invoice $invoice The invoice for the payment
     */
    public function execute_payment_request(nc_payment_invoice $invoice) 
    {
        $invoice_id = $invoice->get_id();
        $order_id = $invoice->get('order_id');
      
        try {
            if ($invoice->get('order_source') !== 'netshop') {
                throw new \Exception('Cannot process invoice with ' . __CLASS__ . ': order_source is not supported');
            }
            
            $nc_core = nc_Core::get_object();
            $netshop = $nc_core->modules->netshop;
            
            $order = $netshop->load_order($order_id);
            if (!$order) {
                throw new \Exception("Order with id={$order_id} not found");
            }
            
            // Return url
            $site = $nc_core->catalogue->get_url_by_id($nc_core->catalogue->id());
            $return_url = $site . 
                          nc_module_path('payment') . 
                          'callback.php?type=result&paySystem=' . __CLASS__ .
                          '&invoice_id=' . $invoice_id;
            
            // Total
            $total_sum = (float) $order->get_totals();
            
            // Customer email
            $customer_email = $invoice->get('customer_email') ?: nc_payment_register::get_default_customer_email();
            
            // Get payment link
            $pay_url = $this->get_pay_url(
                $order_id,
                $customer_email,
                $total_sum,
                $return_url
            );
            
            if (!empty($pay_url)) {
                $invoice->set('last_response', "Payment URL: {$pay_url}");
                $this->pay_url = $pay_url;
            } 
            else {
                $msg = "Failed to get payment URL. Order id: {$order_id}";
                $invoice->set('last_response', $msg)
                        ->set('status', nc_payment_invoice::STATUS_CALLBACK_ERROR);
                $this->add_error($msg);
            }
            
            $invoice->save();
            
        } catch (\Exception $e) {
            // Log the caught exception for debugging purposes.
            $this->add_error(sprintf(
                'VtbPay Exception: %s; Order id: (%s)', 
                $e->getMessage(), 
                $order_id ?: ''
            ));
        }
    }
    
    
    /**
     * Retrieves the payment URL for redirection.
     *
     * @param string $order_id Order ID
     * @param string $email Customer email
     * @param float $total Amount to pay
     * @param string $return_url Return url
     *
     * @return string The payment URL.
     * @throws \Exception Throws an exception if the payment URL cannot be obtained.
     */
    private function get_pay_url(
        string $order_id,
        string $email,
        float $total,
        string $return_url
    ) {
        $response = $this->setRequestData()->getOrderLink(
            $order_id,
            $email,
            time(),
            $total,
            $return_url
        );
        
        $this->vtbpay_logger(
            __FUNCTION__ . ' > getOrderLink', 
            self::LOG_LEVEL_DEBUG, [
            'arguments' => [
                'order_id' => $order_id, 
                'email' => $email,
                'time' => time(), 
                'total' => $total, 
                'return_url' => $return_url
            ],
            'response' => $response
        ]);
        
        return $response['object']['payUrl'] ?? '';
    }
    
    
    /**
     * Validates that all necessary parameters for making a payment request are present
     * Throws an error if any required parameters are missing
     */
    public function validate_payment_request_parameters() 
    {
        if (!$this->get_setting('client_id')) {
            $this->add_error('The "client_id" parameter is required.');
        }

        if (!$this->get_setting('client_secret')) {
            $this->add_error('The "client_secret" parameter is required.');
        }
    }
    
    
    /**
     * Validates the response from the payment callback
     *
     * @param nc_payment_invoice $invoice The invoice associated with the payment
     */
    public function validate_payment_callback_response(nc_payment_invoice $invoice = null) 
    {
        if (!$invoice) {
            $this->add_error(NETCAT_MODULE_PAYMENT_ERROR_INVOICE_NOT_FOUND);
        }
        
        if (empty($this->get_response_value('orderId'))) {
            $this->add_error(NETCAT_MODULE_PAYMENT_ORDER_ID_IS_NULL);
        }
    }
    
    
    /**
     * Handles the response after the payment operation is complete
     *
     * @param nc_payment_invoice $invoice The invoice associated with the payment
     */
    protected function on_response(nc_payment_invoice $invoice = null) 
    {
        $order_id = $this->get_response_value('orderId');
        
        try {
            $payment_status = $this->get_payment_status($order_id);
            
            if ($payment_status !== 'PAID') {
                throw new \Exception('Payment status is not PAID.');
            } 
            
            $this->on_payment_success($invoice);
            $this->response_url = $this->get_setting('success_url') ?: '/';
            
        } catch (\Exception $e) {
            // Log the caught exception for debugging purposes.
            // На данном этапе уже нет смысла вызывать $this->add_error()
            $this->vtbpay_logger(sprintf(
                __FUNCTION__ . ' > VtbPay Exception: %s; Order id: (%s)', 
                $e->getMessage(), 
                $order_id ?: ''
            ), self::LOG_LEVEL_ERROR);
            
            $this->on_payment_failure($invoice);
            $this->response_url = $this->get_setting('fail_url') ?: '/';
        }
    }
    
    
    /**
     * Method requests the payment status from the VtbApi
     * for a specific order and returns it, throwing an exception if no status is received.
     *
     * @param string $order_id Order ID
     *
     * @return string Returns payment status string.
     * @throws \Exception Throws exception when no information about payment status.
     */
    private function get_payment_status(string $order_id)
    {
        $response = $this->setRequestData()->getOrderInfo($order_id);
        
        $this->vtbpay_logger(
            __FUNCTION__ . ' > getOrderInfo', 
            self::LOG_LEVEL_DEBUG, [
            'arguments' => [
                'order_id' => $order_id
            ],
            'response' => $response
        ]);
        
        if (
            !isset($response['object']['status']['value']) ||
            empty($status = $response['object']['status']['value'])
        ) {
            throw new \Exception('No information about payment status.');
        }

        return $status;
    }
    
    
    /**
     * Loads the invoice during the payment callback
     *
     * @return bool|nc_payment_invoice Returns the loaded invoice if it exists, otherwise false
     */
    public function load_invoice_on_callback() 
    {
        return $this->load_invoice($this->get_response_value('invoice_id'));
    }
    
    
    /**
     * Logs an error and adds it to the list of errors
     *
     * @param string $string The error message to be logged and added
     */
    protected function add_error($string) 
    {
        parent::add_error($string);
        
        $this->vtbpay_logger($string, self::LOG_LEVEL_ERROR);
    }
    
    
    /**
     * VtbApi
     */
     
    /**
     * Sets up the necessary request data for the VTB payment API
     *
     * @return $this Returns the current instance for method chaining
     */
    private function setRequestData() 
    {
        $this->clientData['client_id'] = trim($this->get_setting('client_id'));
        $this->clientData['client_secret'] = trim($this->get_setting('client_secret'));
        $this->clientMerchant = trim($this->get_setting('merchant_authorization') ?: '');
        $this->mode = ($this->get_setting('test_mode') == 1) ? 'TEST': 'PROD';

        $this->tokenUrl = self::$modeUrls[$this->mode]['TOKEN'];
        $this->orderUrl = self::$modeUrls[$this->mode]['ORDER'];
        
        return $this;
    }
    

    /**
     * Retrieves the access token for the VTB payment API
     *
     * @return mixed The response from the API, which should include the access token
     */
    private function getAccessToken() 
    {
        $headers = [
            'Content-Type: application/x-www-form-urlencoded'
        ];
        if (!empty($merchant = $this->clientMerchant)) {
            $headers[] = 'Merchant-Authorization: ' . $merchant;
        }
        
        return $this->sendRequest(
            $this->tokenUrl,
            http_build_query($this->clientData),
            $headers
        );
    }
    

    /**
     * Creates an order link for the specified parameters
     *
     * @param string $orderId The ID of the order
     * @param string $email The email of the customer
     * @param int $orderCreate The creation timestamp of the order
     * @param float $amount The amount of the order
     * @param string $returnUrl The URL to redirect to after the order
     * @return mixed The result from the sendRequest method, or false if the headers or merchant data is missing
     */
    private function getOrderLink(
        string $orderId,
        string $email,
        int $orderCreate,
        float $amount,
        string $returnUrl
    ) {
        $headers = $this->getAutorizationHeaders();

        $headers[] = 'Content-Type: application/json';
        if (!empty($merchant = $this->clientMerchant)) {
            $headers[] = 'Merchant-Authorization: ' . $merchant;
        }

        $timestamp = $orderCreate + 86400; // +1d
        $expire = date('Y-m-d\TH:i:s.v\Z', $timestamp); // ISO 8601

        $args = [
            'orderId' => $orderId,
            'orderName' => 'Заказ #' . $orderId,
            'customer' => [
                'email' => $email
            ],
            'amount' => [
                'value' => $amount,
                'code' => 'RUB'
            ],
            'expire' => $expire,
            'returnUrl' => $returnUrl
        ];

        return $this->sendRequest(
            $this->orderUrl,
            $args,
            $headers
        );
    }
    

    /**
     * Retrieves the information for a specific order
     *
     * @param string $orderId The ID of the order
     * @return mixed The result from the sendRequest method, or false if the headers or merchant data is missing
     */
    private function getOrderInfo(string $orderId)
    {
        $headers = $this->getAutorizationHeaders();

        if (!empty($merchant = $this->clientMerchant)) {
            $headers[] = 'Merchant-Authorization: ' . $merchant;
        }

        return $this->sendRequest(
            $this->orderUrl . '/' . $orderId,
            [],
            $headers,
            'GET'
        );
    }
    

    /**
     * Generates the authorization headers for the API request
     *
     * @return array|bool The authorization headers if the access token exists, otherwise false
     */
    private function getAutorizationHeaders()
    {
        $token = $this->getAccessToken();
        
        if(!isset($token['access_token'])) {
          
            $errorDescription = '';
            if (
                isset($token['error_description']) && 
                !empty($token['error_description'])
            ) $errorDescription .= ' Error description: ' . $token['error_description'];
            
            throw new \Exception('Error getting access token.' . $errorDescription);
        }

        return [
            'X-IBM-Client-Id: ' . explode('@', $this->clientData['client_id'])[0],
            'Authorization: ' . $token['token_type'] . ' ' . $token['access_token'],
        ];
    }
    
    
    /**
     * Sends a HTTP request to the provided URL
     *
     * @param string $url The URL to send the request to
     * @param array $data The data to send with the request
     * @param array $headers The headers to send with the request
     * @param string $method The HTTP method to use for the request (default 'POST')
     * @param bool $json_encode Whether or not to JSON encode the data (default false)
     * @return mixed The result of the request, decoded from JSON, or false if there was an error with the request
     */
    private function sendRequest(
        $url,
        $data = [],
        $headers = [],
        $method = 'POST',
        $jsonEncode = true
    ) {
        $curl = curl_init();
        
        if (in_array('Content-Type: application/json', $headers) && $jsonEncode) {
            $data = json_encode($data);
        }

        if ($method === 'POST') {
            curl_setopt($curl, CURLOPT_POST, 1);
            curl_setopt($curl, CURLOPT_POSTFIELDS, $data);
        } 
        elseif ($method === 'PUT') {
            curl_setopt($curl, CURLOPT_CUSTOMREQUEST, 'PUT');
            curl_setopt($curl, CURLOPT_POSTFIELDS, $data);
        } 
        else {
            $url .= is_array($data) ? '?' . http_build_query($data) : '';
        }

        curl_setopt($curl, CURLOPT_URL, $url);
        curl_setopt($curl, CURLOPT_HTTPHEADER, $headers);
        curl_setopt($curl, CURLOPT_RETURNTRANSFER, true);
        curl_setopt($curl, CURLOPT_SSL_VERIFYPEER, false);

        $result = curl_exec($curl);
        
        if ($result === false) {
            throw new \Exception('Curl error: ' . curl_error($curl));
        }
        
        if (empty($result)) {
            throw new \Exception('The curl request returned an empty response');
        }
        
        $result = json_decode($result, true);
        
        if (json_last_error() != JSON_ERROR_NONE) {
            throw new \Exception('Json decode error: ' . json_last_error_msg());
        }

        curl_close($curl);

        return $result;
    }
    
    
    /**
     * Logs a message with the provided context
     *
     * @param string $message The message to log
     * @param string $type The type of log (default 'DEBUG')
     * @param array $context The context to log with the message
     * @param bool $print_r Whether or not to use print_r instead of json_encode for the context (default false)
     * @return bool Whether or not the log was successful
     */
    private function vtbpay_logger(
        string $message, 
        string $type = self::LOG_LEVEL_DEBUG, 
        array $context = [],
        bool $print_r = false
    ) {
        if (
            empty($message) || 
            !$this->get_setting('logging')
        ) return null;
        
        $message = '[' . date('Y-m-d H:i:s') . '] ' . $type . ' ' . __CLASS__ . ' - ' . $message;

        if (!empty($context)) {
            $message .= ' Context: ' . PHP_EOL;
            $message .= $print_r ? print_r($context, true) : json_encode($context, JSON_UNESCAPED_UNICODE);
        }
        
        $dir = $_SERVER['DOCUMENT_ROOT'] . '/netcat_files/logs';
        
        if (!is_dir($dir)) {
            mkdir($dir, 0777, true);
        }
        
        error_log($message . PHP_EOL, 3, $dir . '/vtbpay.log');
        
    }
}